import pytest

from configstore.converters import asbool


@pytest.mark.parametrize('value', [
    "true", " true ", " TrUe",
    "on", "ON",
    "\nYes", " \tyes ",
    "1", " 1 ", 1,
    True,
])
def test_asbool_true(value):
    assert asbool(value) is True


@pytest.mark.parametrize('value', [
    "False", "FALSE", "false  ",
    "off", "oFF ",
    "no", "\tno",
    " 0 ", "0", 0,
    "", " ", "\n", "\t",
    False,
])
def test_asbool_false(value):
    assert asbool(value) is False


@pytest.mark.parametrize('value', [
    "y", "n", "t", "f",
    "00", "01", None,
])
def test_asbool_invalid(value):
    with pytest.raises(ValueError):
        asbool(value)



class DictBackend:
    """Simple backend to populate settings from a dictionary."""

    def __init__(self, settings):
        self._settings = settings
        # we're not making a copy or deep copy here; intended usage is
        # to create a dict on the fly (like DictBackend({"key": value})).
        # passing a dict that survives and can be mutated is possible.

    def get_setting(self, key):
        return self._settings.get(key)



import pytest

from configstore import Store, SettingNotFoundException, DictBackend


def test_store_init():
    Store([])
    Store([DictBackend({})])

    with pytest.raises(Exception):
        Store()
    with pytest.raises(Exception):
        Store(None)


def test_store_init_invalid():
    with pytest.raises(TypeError):
        Store([DictBackend])


def test_store_get_setting():
    store = Store([DictBackend({'key': 'secrets!'})])

    value = store.get_setting('key')

    assert value == 'secrets!'


def test_store_get_setting_with_default():
    store = Store([DictBackend({'key': 'secrets!'})])

    value = store.get_setting('key', 'default')

    assert value == 'secrets!'


def test_store_get_setting_missing():
    store = Store([])

    with pytest.raises(SettingNotFoundException):
        store.get_setting('key')


def test_store_get_setting_missing_with_default():
    store = Store([DictBackend({})])

    value = store.get_setting('key', 'default value')

    assert value == 'default value'


def test_store_interpolate():
    store = Store([DictBackend({'environment': 'staging'})])

    s = store.interpolate('before ${environment} after')

    assert s == 'before staging after'


def test_store_interpolate_none_default():
    store = Store([])

    assert store.get_setting('foo', None) is None


def test_store_get_setting_interpolate_value():
    store = Store([DictBackend(dict(
        environment='staging',
        secret_key='42-${environment}-secrets!',
    ))])

    value = store.get_setting('secret_key')

    assert value == '42-staging-secrets!'


def test_store_get_setting_interpolate_default():
    store = Store([DictBackend({'service_host': 'cool-db-server:6000', 'db': '42'})])

    value = store.get_setting('service_url', 'https://${service_host}/${db}')

    assert value == 'https://cool-db-server:6000/42'


def test_store_add_backend():
    store = Store([])

    store.add_backend(DictBackend({'environment': 'staging'}))

    assert store.get_setting('environment') == 'staging'


def test_store_add_backend_invalid():
    class BadBackend:
        def get_key(key):
            """oops wrong name"""

    store = Store([])
    bad = BadBackend()

    with pytest.raises(TypeError):
        store.add_backend(bad)



import os


class EnvVarBackend:
    """Backend that reads settings in environment variables."""

    def __init__(self, name_prefix=''):
        self.name_prefix = ''

    def get_setting(self, key):
        key = self.name_prefix + key
        return os.environ.get(key)



import configstore

# Create a store object with a stack of backends to look for settings
# in environment first, then docker secrets, then a .env file
store = configstore.Store([
    configstore.EnvVarBackend(),
    configstore.DockerSecretBackend(),
])

# Add backend to read Docket secrets
configstore.DotenvBackend('/path/to/env/file'),

# Will raise an exception if APP_SECRET_KEY isn't found in any backend
SECRET_KEY = store.get_setting('APP_SECRET_KEY')

# Will use 'defaultvalue' if APP_SECRET_KEY isn't found
SECRET_KEY = store.get_setting('APP_SECRET_KEY', 'defaultvalue')

# Setting values and default values can reference other settings
BROKER_URL = store.get_setting('BROKER_URL', '${REDIS_URL}/7')

# Helper param for bool strings (true, false, on, off, yes, no, 1, 0)
DEBUG_MODE = store.get_setting('DEBUG_MODE', 'off', asbool=True)

# Example for list/tuple/etc
# setting value is whitespace-separated: "bob  alice \n marcel"
ADMINS = store.get_setting('ADMINS').split()

del store



import pytest
import configstore

DOTENV_CONTENTS = u'''
SECRET_KEY=1234dot
ENVIRONMENT=PRODUCTION
EMAIL_DEBUG=1
PROMO2=off
'''

DOTENV_EMPTY = u'''

'''


def test_envvars_override_dotenv(monkeypatch, tmp_path):
    monkeypatch.setenv('ENVIRONMENT', 'STAGING')
    path = tmp_path / 'config.env'
    path.write_text(DOTENV_CONTENTS)

    store = configstore.Store([
        configstore.EnvVarBackend(),
        configstore.DotenvBackend(str(path)),
    ])

    secret_key = store.get_setting('SECRET_KEY')
    environment = store.get_setting('ENVIRONMENT')

    assert secret_key == '1234dot'
    assert environment == 'STAGING'


def test_dotenv_overrides_envvars(monkeypatch, tmp_path):
    monkeypatch.setenv('ENVIRONMENT', 'STAGING')
    path = tmp_path / 'config.env'
    path.write_text(DOTENV_CONTENTS)

    store = configstore.Store([
        configstore.DotenvBackend(str(path)),
        configstore.EnvVarBackend(),
    ])

    secret_key = store.get_setting('SECRET_KEY')
    environment = store.get_setting('ENVIRONMENT')

    assert secret_key == '1234dot'
    assert environment == 'PRODUCTION'


def test_dotenv_empty_file(monkeypatch, tmp_path):
    monkeypatch.setenv('ENVIRONMENT', 'STAGING')
    path = tmp_path / 'config.env'
    path.write_text(DOTENV_EMPTY)

    store = configstore.Store([
        configstore.DotenvBackend(str(path)),
        configstore.EnvVarBackend(),
    ])

    with pytest.raises(Exception) as excinfo:
        store.get_setting('DEBUG')
    environment = store.get_setting('ENVIRONMENT')

    assert "Couldn't find setting" in str(excinfo.value)
    assert environment == 'STAGING'


def test_dict_defaults(monkeypatch):
    defaults = {
        'DEBUG': False,
        'TIMEOUT': 100,
    }
    monkeypatch.setenv('TIMEOUT', '5')
    monkeypatch.setenv('EMAIL', 'smtp://localhost')

    store = configstore.Store([
        configstore.EnvVarBackend(),
        configstore.DictBackend(defaults),
    ])

    debug = store.get_setting('DEBUG')
    timeout = int(store.get_setting('TIMEOUT'))
    email = store.get_setting('EMAIL')

    assert debug is False
    assert timeout == 5
    assert email == 'smtp://localhost'


def test_interpolation(monkeypatch, tmp_path):
    monkeypatch.setenv('ENVIRONMENT', 'STAGING')
    monkeypatch.setenv('VERSION', '113')
    monkeypatch.setenv('REDIS_URL', 'https://redis:4012')
    path = tmp_path / 'config.env'
    path.write_text(DOTENV_CONTENTS)
    path.write_text('APP_NAME=supertest-${ENVIRONMENT}-${VERSION}')

    store = configstore.Store([
        configstore.EnvVarBackend(),
        configstore.DotenvBackend(str(path)),
    ])

    environment = store.get_setting('ENVIRONMENT')
    app_name = store.get_setting('APP_NAME')
    session_url = store.get_setting('SESSION_URL', '${REDIS_URL}/2')

    assert environment == 'STAGING'
    assert app_name == 'supertest-STAGING-113'
    assert session_url == 'https://redis:4012/2'


def test_conversion(monkeypatch, tmp_path):
    monkeypatch.setenv('DEBUG', 'on')
    path = tmp_path / 'config.env'
    path.write_text(DOTENV_CONTENTS)

    store = configstore.Store([
        configstore.DotenvBackend(str(path)),
        configstore.EnvVarBackend(),
    ])

    # basic test
    email_as_string = store.get_setting('EMAIL_DEBUG')
    email_as_bool = store.get_setting('EMAIL_DEBUG', asbool=True)

    assert email_as_string == '1'
    assert email_as_bool is True

    # setting found in store, value true
    debug = store.get_setting('DEBUG', asbool=True)
    # not found, default is boolean
    promo1 = store.get_setting('PROMO1', True, asbool=True)
    # found, value false
    promo2 = store.get_setting('PROMO2', asbool=True)
    # not found, default converted from string
    promo3 = store.get_setting('PROMO3', '1', asbool=True)
    # not found, default w/ interpolation, false
    promo3 = store.get_setting('PROMO3', '${PROMO2}', asbool=True)
    # not found, default w/ interpolation, true
    extra_debug = store.get_setting('EXTRA_DEBUG', '${DEBUG}', asbool=True)

    assert debug is True
    assert promo1 is True
    assert promo2 is False
    assert promo3 is False
    assert extra_debug is True

    # error in conversion bubbles up
    with pytest.raises(ValueError):
        store.get_setting('ENVIRONMENT', asbool=True)
